# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '14:12'


class BaseNosqlConnector(ABC):
    def __init__(self):
        self.tagprefix = "tag:"

    @abstractmethod
    def delete_table(self):
        """
        deletes table from db
        :return:
        """

    @abstractmethod
    def create_table(self):
        """
        creates a new table, delete first if exists \n
        :return:
        """
    @abstractmethod
    def dbsize(self) -> int:
        """
        returns the itemcount in db
        :return: item count
        """

    @abstractmethod
    def keys(self, pattern='*') -> list:
        """
        :param pattern: str urlpathregex
        :return: the keys in db
        """

    @abstractmethod
    def get(self, key: str) -> str or dict or list:
        """
        get value by key\n
        :param key: str
        :return: str or dict or list
        """

    @abstractmethod
    def get_all_as_dict(self) -> dict:
        """
        get all values in store \n
        :return:dict
        """

    @abstractmethod
    def get_all_as_list(self) -> list:
        """
        get all values in store \n
        :return:dict
        """

    @abstractmethod
    def upsert(self, key: str, value: str or list or dict) -> bool:
        """
        adds a value to store
        :param key: key
        :param value: value to insert
        :return: bool
        """

    @abstractmethod
    def remove(self, key: str) -> bool:
        """
        removes the key from store
        :param key: str or list or tuple
        :return:
        """

    @abstractmethod
    def remove_keys(self, keys: list) -> bool:
        """
        removes the key from store
        :param keys:
        :return:
        """

    @abstractmethod
    def flush(self):
        """
        Flushes current store\n
        :return: None
        """

    @abstractmethod
    def tags(self) -> list:
        """
        returns tag list \n
        :return:
        """

    @abstractmethod
    def gettagkeys(self, tag: str) -> list:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """

    @abstractmethod
    def gettag(self, tag: str) -> list:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """

    @abstractmethod
    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """

    @abstractmethod
    def removefromtag(self, tag: str, key: str or list):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """



# -*- coding: utf-8 -*-
"""
Basic redis connector
"""
import json

import redis

from nosqlmodel.connectors.baseconnector import BaseNosqlConnector

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '10:07'

RHOST = "localhost"
RPORT = "6379"


class RedisConnector(BaseNosqlConnector):

    def __init__(self, dbnum: int = 0):
        self.redisdbnum = dbnum
        self.conn = redis.StrictRedis(host=RHOST, port=RPORT, charset="utf-8",
                                      decode_responses=True, db=self.redisdbnum)
        super().__init__()

    def delete_table(self):
        self.flush()

    def create_table(self):
        pass

    def dbsize(self) -> int:
        """
        returns the itemcount in db
        :return: item count
        """
        # return self.conn.dbsize()
        return len(self.keys())

    def keys(self, pattern='*') -> list:
        """
        :param pattern: str urlpathregex
        :return: the keys in db
        """
        return list(
            set(self.conn.keys(pattern=pattern)) - set(self.conn.keys(self.tagprefix + "*")))

    def get(self, key: str) -> str or dict or list:
        """
        get value by key\n
        :param key: str
        :return: str or dict or list
        """
        key = str(key)
        retval = self.conn.get(key)

        if retval:
            # noinspection PyBroadException
            try:
                retval = json.loads(retval, encoding="UTF-8")
            except Exception:
                retval = str(retval)
        else:
            pass

        return retval

    def get_all_as_dict(self) -> dict:
        """
        get all values in store \n
        :return:dict
        """
        keys = self.keys()
        retdct = {}
        for key in keys:
            retdct[key] = self.get(key)

        return retdct

    def get_all_as_list(self) -> list:
        """
        get all values in store \n
        :return:dict
        """
        keys = self.keys()
        retlist = []
        for key in keys:
            retlist.append(self.get(key))

        return retlist

    def upsert(self, key: str, value: str or list or dict) -> bool:
        """
        adds a value to store
        :param key: key
        :param value: value to insert
        :return: bool
        """
        key = str(key)
        if isinstance(value, (list, (dict, set, tuple))):
            value = json.dumps(value, ensure_ascii=False)
        elif isinstance(value, object) and hasattr(value, "__dict__"):
            value = json.dumps(value.__dict__, ensure_ascii=False)
        else:
            value = str(value)
        ret = self.conn.set(key, value)
        if ret:
            return True
        else:
            return False

    def remove(self, key: str) -> bool:
        """
        removes the key from store
        :param key: str or list or tuple
        :return:
        """
        key = str(key)
        ret = str(self.conn.delete(key))
        return {"1": True}.get(ret, False)

    def remove_keys(self, keys: list) -> bool:
        """
        removes the key from store
        :param keys:
        :return:
        """
        ret = str(self.conn.delete(keys))
        return {"1": True}.get(ret, False)

    def flush(self):
        """
        Flushes current store\n
        :return: None
        """
        self.conn.flushdb()

    def tags(self) -> list:
        """
        returns tag list \n
        :return:
        """
        taglist = self.conn.keys(self.tagprefix + "*")
        retlist = []
        for tag in taglist:
            retlist.append(tag.replace(self.tagprefix, ""))
        return retlist

    def gettagkeys(self, tag: str) -> list:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """
        memberids = self.conn.smembers(self.tagprefix + tag)
        return list(memberids)

    def gettag(self, tag: str) -> list:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """
        memberids = self.gettagkeys(tag)
        retlist = []
        for memberid in list(memberids):
            retlist.append(self.get(memberid))

        return retlist

    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """
        if not (tag and key):
            raise TypeError
        self.conn.sadd(self.tagprefix + tag, key)

    def removefromtag(self, tag: str, key: str or list):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """

        tag = str(tag)
        # ! dont touch it. srem doesnt accept list !!!!!
        if isinstance(key, (list, tuple, set)):
            for k in key:
                self.conn.srem(self.tagprefix + tag, k)
        else:
            self.conn.srem(self.tagprefix + tag, key)



# -*- coding: utf-8 -*-
import json
import os
import uuid
import zipfile

from simplejsonobject import JsonObject

from nosqlmodel.connectors.redisconnector import RedisConnector

__author__ = 'ozgur'
__creation_date__ = '12.09.2019' '13:48'


class BaseNoSqlModel(JsonObject):
    """
    Nosql model with basic functionality\n
    """

    class Meta:
        connector = RedisConnector()
        indexkey = None

    def __init__(self):
        self.idkey = None

    def delete_table(self):
        """
        deletes table from db
        :return:
        """
        self.Meta.connector.delete_table()

    def create_table(self):
        """
        creates a new table, delete first if exists \n
        :return:
        """
        self.Meta.connector.create_table()

    @classmethod
    def flush(cls):
        """
        flushes al records
        :return: None
        """
        cls.Meta.connector.flush()

    @classmethod
    def dbsize(cls) -> int:
        """
        returns the itemcount in db
        :return: item count
        """
        return cls.Meta.connector.dbsize()

    @classmethod
    def get_keys(cls, pattern: str = "*") -> list:
        """
        return all idkeys in db
        :return:
        """
        return cls.Meta.connector.keys(pattern)

    def get_by_id(self, idkey: str) -> bool:
        """
        get item by id
        :return:
        """
        ret = self.Meta.connector.get(idkey)
        if ret:
            self.from_dict(ret)
            return True
        else:
            return False

    def get_or_create_by_id(self, idkey: str) -> bool:
        """
        get item by id
        :return:
        """
        ret = self.Meta.connector.get(idkey)
        if ret:
            self.from_dict(ret)
            return True
        else:
            self.__dict__[self.Meta.indexkey] = idkey
            return False

    @classmethod
    def get_all_as_list(cls) -> list:
        """
        returns all db as list of item dicts\n
        :return:  [object_dict]
        """
        return cls.Meta.connector.get_all_as_list()

    @classmethod
    def get_all_as_dict(cls) -> dict:
        """
        returns all db as dict of item dicts\n
        :return:  {"id":object_dict,}
        """
        return cls.Meta.connector.get_all_as_dict()

    @classmethod
    def get_all_as_objectlist(cls) -> list:
        """
        returns all db as list of item objects\n
        :return:  [object]
        """
        retlist = []
        for itemdict in cls.get_all_as_list():
            retobj = cls()
            retobj.from_dict(itemdict)
            retlist.append(retobj)
        return retlist

    @classmethod
    def get_all_as_objectdict(cls) -> dict:
        """
        returns all db as dict of item objects\n
        :return:  {key:object}
        """
        retdict = {}
        for key, itemdict in cls.get_all_as_dict():
            retobj = cls()
            retobj.from_dict(itemdict)
            retdict[key] = retobj
        return retdict

    def save_to_cache(self, tags: list = None) -> bool:
        """
        saves object
        :return: success status
        """
        if not self.idkey:
            if self.Meta.indexkey:
                bnoid = self.__dict__.get(self.Meta.indexkey, None)
                self.idkey = bnoid if bnoid else str(uuid.uuid4())
            else:
                self.idkey = str(uuid.uuid4())
        else:
            pass
        retval = self.Meta.connector.upsert(self.idkey, self.to_dict())

        if tags:
            tnew = [x for x in tags if x]
            self.add_tags_to_item(tnew)

        return retval

    def delete(self):
        """
        Deletes item from redis \n
        :return: None
        """
        if not self.idkey:
            raise KeyError("First you must save to remove it")
        self.Meta.connector.remove(self.idkey)

    @classmethod
    def get_tag_keys(cls, tag: str) -> list:
        """
        returns keys tagged bay given tag
        :return: list
        """
        return cls.Meta.connector.gettagkeys(tag)

    @classmethod
    def get_tags(cls) -> list:
        """
        returns tag list
        :return: list
        """
        return cls.Meta.connector.tags()

    @classmethod
    def get_by_tag(cls, tag: str) -> list:
        """
        returns obejct list tagged by given tag
        :return:[BaseNosqlModel]
        """
        retlist = []
        for itemdict in cls.Meta.connector.gettag(tag):
            retobj = cls()
            retobj.from_dict(itemdict)
            retlist.append(retobj)
        return retlist

    def add_tags_to_item(self, taglist: list):
        """
        adds tags to object
        :param taglist:
        :return:
        """
        if not self.idkey:
            raise KeyError("You must save before adding a tag")
        for tag in taglist:
            self.Meta.connector.addtag(tag, self.idkey)

    def remove_item_from_tag(self, tag: str):
        if not self.idkey:
            raise KeyError("You must save before removing from a tag")
        self.Meta.connector.removefromtag(tag, self.idkey)

    def from_dict(self, updatedict: dict):
        """
        populates data from dict
        Warning this will also overrides the id too!!        \n
        :param updatedict: dict which contains data
        :return: None
        """
        try:
            del updatedict["connector"]
        except (KeyError, TypeError):
            pass
        super().from_dict(updatedict.copy())

    def to_dict(self) -> dict:
        """
        populates data to dict \n
        :return: None
        """
        retval = super().to_dict()
        try:
            del retval["connector"]
        except (KeyError, TypeError):
            pass
        return retval

    @classmethod
    def export_to_json_text(cls, exportdict: dict = None, compress_data=False) -> str:
        """
         transforms exportdict or whole database into json \n
        :param compress_data: bool, data will be compressed or not
        :param exportdict: must be dictionary of same class objects
        :return: returns a json compliant text file
        """
        edict = {}
        if exportdict:
            for key, value in exportdict.items():
                try:
                    edict[key] = value.to_dict()
                except AttributeError:
                    pass
        else:
            edict = cls.Meta.connector.get_all_as_dict()
        if compress_data:
            for key, subdict in edict.items():
                edict[key] = dict((i, d) for i, d in subdict if d)

            retval = json.dumps(edict, ensure_ascii=False, indent=4)
        else:
            retval = json.dumps(edict, ensure_ascii=False)
        return retval

    @classmethod
    def export_to_json_file(cls, filepath: str, exportdict: dict = None, compress_data=False):
        """
         transforms exportdict or whole database into *.json file \n
        :param filepath: must end with .json
        :param exportdict: must be dictionary of same class objects
        :param compress_data: bool, data will be compressed or not
        :return:
        """
        with open(filepath, "w") as ofile:
            ofile.write(cls.export_to_json_text(exportdict, compress_data=compress_data))

    @classmethod
    def export_to_json_zip(cls, filepath: str, exportdict: dict = None, compress_data=False):
        """
         transforms exportdict or whole database into *.zip file \n
        :param filepath: must be end with .zip
        :param exportdict: must be dictionary of same class objects
        :param compress_data: bool, data will be compressed or not
        :return:
        """
        if not filepath.endswith(".zip"):
            raise FileNotFoundError("Wrong filename:" + filepath)
        else:
            tempfilepath = filepath.replace(".zip", ".json")
        cls.export_to_json_file(tempfilepath, exportdict, compress_data=compress_data)
        zipfile.ZipFile(filepath, mode='w').write(tempfilepath)
        os.remove(tempfilepath)

    @classmethod
    def import_from_json_text(cls, jsontext: str):
        """
        import and update db from exported json text \n
        :param jsontext: exported json text
        :return:
        """
        datadict = json.loads(jsontext, encoding="UTF-8")
        for key, value in datadict.items():
            instance = cls()
            instance.from_dict(value)
            instance.save_to_cache()

    @classmethod
    def import_from_json_file(cls, filepath: str):
        """
        import and update db from exported *.json file \n
        :param filepath: exported *.json file
        :return:
        """
        with open(filepath, "r") as ofile:
            content = ofile.read()
            cls.import_from_json_text(content)

    @classmethod
    def import_from_json_zip(cls, filepath: str):
        """
        import and update db from exported *.json file \n
        :param filepath: exported *.json file
        :return:
        """
        if not filepath.endswith(".zip"):
            raise FileNotFoundError("Wrong filename:" + filepath)
        else:
            tempfilepath = filepath.replace(".zip", ".json")

        with zipfile.ZipFile(filepath, 'r') as zipObj:
            zipObj.extractall("/")
        cls.import_from_json_file(tempfilepath)
        os.remove(tempfilepath)



# -*- coding: utf-8 -*-
"""
Basic dynamodb connector
"""
import json
import time
from decimal import Decimal

import boto3
# from boto3.resources.factory.dynamodb import Table
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError

from nosqlmodel.connectors.baseconnector import BaseNosqlConnector

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '10:07'

D_REGION_NAME = ""
D_AWS_ACCESS_KEY_ID = ""
D_AWS_SECRET_ACCESS_KEY = ""


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, Decimal):
            if abs(o) % 1 > 0:
                return float(o)
            else:
                return int(o)
        return super(DecimalEncoder, self).default(o)


class DynamoConnector(BaseNosqlConnector):

    def __init__(self, tablename: str):
        self.source = boto3.resource(
            'dynamodb',
            region_name=D_REGION_NAME,
            aws_access_key_id=D_AWS_ACCESS_KEY_ID,
            aws_secret_access_key=D_AWS_SECRET_ACCESS_KEY
        )
        self.table = self.source.Table(tablename)  # pylint: disable=E1101
        self.taglistkey = "tags"
        super().__init__()

    @staticmethod
    def _f_to_d(fval: float) -> Decimal:
        """
        float to decimal converter \n
        :param fval: float value
        :return: Decimal
        """
        return Decimal(str(fval))

    @staticmethod
    def _d_to_f(dval: Decimal) -> float:
        """
        Deciaml to float converter \n
        :param dval: Decimal value
        :return: float
        """
        return float(dval)

    def delete_table(self):
        try:
            self.table.delete()
            time.sleep(10)
        except ClientError:
            pass

    def create_table(self):
        self.table = self.source.create_table(  # pylint: disable=E1101
            TableName='Test',
            KeySchema=[
                {
                    'AttributeName': 'idkey',
                    'KeyType': 'HASH'  # Partition key
                },
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'idkey',
                    'AttributeType': 'S'
                },

            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 10
            }
        )
        time.sleep(10)

    @staticmethod
    def issuccess(resp: dict) -> bool:
        """
        checks if request is success (returns 200)
        :param resp: boto response dict
        :return: bool
        """
        if resp.get("ResponseMetadata", {}).get("HTTPStatusCode", 0) is 200:
            return True
        else:
            return False

    def keys(self, pattern='*') -> list:
        """
        :param pattern: str urlpathregex
        :return: the keys in db
        """
        retlist = []
        resp = self.table.scan(
            ProjectionExpression='idkey',
        )
        data = resp['Items']

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                ExclusiveStartKey=resp['LastEvaluatedKey'],
                ProjectionExpression='idkey',
            )
            data.extend(resp['Items'])

        for rdict in data:
            val = rdict['idkey']
            if not str(val).startswith(self.tagprefix):
                retlist.append(val)
        return retlist

    def dbsize(self) -> int:
        """
        returns the itemcount in db
        :return: item count
        """
        # return self.conn.dbsize()
        return len(self.keys())

    def get(self, key: str) -> str or dict or list:
        """
        get value by key\n
        :param key: str
        :return: str or dict or list
        """
        key = str(key)
        retval = self.table.get_item(
            Key={
                'idkey': key
            }
        ).get("Item", None)
        if retval:
            for key, value in retval.items():
                if isinstance(value, Decimal):
                    retval[key] = self._d_to_f(value)

        return retval

    def get_all_as_list(self) -> list:
        """
        get all values in store \n
        :return:dict
        """

        resp = self.table.scan()
        data = resp['Items']
        tagindexlist = []

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                ExclusiveStartKey=resp['LastEvaluatedKey']
            )
            data.extend(resp['Items'])
        for index, ddict in enumerate(data):
            if str(ddict["idkey"]).startswith(self.tagprefix):
                tagindexlist.append(index)
            else:
                for key, value in ddict.items():
                    if isinstance(value, Decimal):
                        ddict[key] = self._d_to_f(value)
                data[index] = ddict

        tagindexlist.reverse()
        for index in tagindexlist:
            del data[index]

        return data

    def get_all_as_dict(self) -> dict:
        """
        get all values in store \n
        :return:dict
        """
        retdict = {}
        plist = self.get_all_as_list()
        for idict in plist:
            retdict[idict["idkey"]] = idict
        return retdict

    def upsert(self, key: str, value: dict) -> bool:
        """
        adds a value to store
        :param key: key
        :param value: value to insert
        :return: bool
        """
        value = value.copy()
        if "idkey" not in value:
            value["idkey"] = key

        for key, data in value.items():
            if isinstance(data, float):
                value[key] = self._f_to_d(data)
        resp = self.table.put_item(Item=value)
        return self.issuccess(resp)

    def remove(self, key: str) -> bool:
        """
        removes the key from store
        :param key: str or list or tuple
        :return:
        """

        resp = self.table.delete_item(
            Key={
                'idkey': str(key),
            }
        )
        return self.issuccess(resp)

    def remove_keys(self, keys: list) -> bool:
        """
        removes the key from store
        :param keys:
        :return:
        """
        # ret = str(self.table.delete(keys))
        # return {"1": True}.get(ret, False)

        with self.table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(Key={'idkey': str(key)})
        return True

    def flush(self):
        """
        Flushes current store\n
        :return: None
        """
        keys = self.keys()
        with self.table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(Key={'idkey': key})

    def tags(self) -> list:
        """
        returns tag list \n
        :return:
        """
        fe = Key("idkey").begins_with(self.tagprefix)
        resp = self.table.scan(
            FilterExpression=fe,
            ProjectionExpression='idkey',
        )
        data = resp['Items']

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                FilterExpression=fe,
                ProjectionExpression='idkey',
                ExclusiveStartKey=resp['LastEvaluatedKey']
            )
            data.extend(resp['Items'])
        for index, ddict in enumerate(data):
            data[index] = ddict

        return [rdict['idkey'].replace(self.tagprefix, "", 1) for rdict in data]

    def gettagkeys(self, tag: str) -> list:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """
        tdict = self.get(self.tagprefix + tag)
        retval = tdict[self.taglistkey] if tdict else []
        return retval

    def gettag(self, tag: str) -> list:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """
        memberids = self.gettagkeys(tag)
        retlist = []
        for memberid in list(memberids):
            retlist.append(self.get(memberid))

        return retlist

    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """

        if not (tag and key):
            raise TypeError
        tags = self.gettagkeys(tag)
        tags.append(key)
        tags = list(set(tags))
        ukey = self.tagprefix + tag
        self.upsert(ukey, {"idkey": ukey, self.taglistkey: tags})

    def removefromtag(self, tag: str, key: str or list):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """

        tags = list(set(self.gettagkeys(tag)))
        if isinstance(key, str):
            # noinspection PyBroadException
            try:
                tags.remove(key)
            except Exception:
                pass
        else:
            for k in key:
                # noinspection PyBroadException
                try:
                    tags.remove(k)
                except Exception:
                    pass

        ukey = self.tagprefix + tag
        self.upsert(ukey, {"idkey": ukey, self.taglistkey: tags})
